package sun.util.locale.provider;

import java.util.ArrayList;
import java.util.Stack;
import java.text.CharacterIterator;
import java.io.IOException;

class DictionaryBasedBreakIterator extends RuleBasedBreakIterator
{
    private BreakDictionary dictionary;
    private boolean[] categoryFlags;
    private int dictionaryCharCount;
    private int[] cachedBreakPositions;
    private int positionInCache;
    
    DictionaryBasedBreakIterator(String s, String s2) throws IOException {
        super(s);
        byte[] additionalData = super.getAdditionalData();
        if (additionalData != null) {
            this.prepareCategoryFlags(additionalData);
            super.setAdditionalData((byte[])null);
        }
        this.dictionary = new BreakDictionary(s2);
    }
    
    private void prepareCategoryFlags(byte[] array) {
        this.categoryFlags = new boolean[array.length];
        for (int i = 0; i < array.length; ++i) {
            this.categoryFlags[i] = (array[i] == 1);
        }
    }
    
    @Override
    public void setText(CharacterIterator text) {
        super.setText(text);
        this.cachedBreakPositions = null;
        this.dictionaryCharCount = 0;
        this.positionInCache = 0;
    }
    
    @Override
    public int first() {
        this.cachedBreakPositions = null;
        this.dictionaryCharCount = 0;
        this.positionInCache = 0;
        return super.first();
    }
    
    @Override
    public int last() {
        this.cachedBreakPositions = null;
        this.dictionaryCharCount = 0;
        this.positionInCache = 0;
        return super.last();
    }
    
    @Override
    public int previous() {
        CharacterIterator text = this.getText();
        if (this.cachedBreakPositions != null && this.positionInCache > 0) {
            --this.positionInCache;
            text.setIndex(this.cachedBreakPositions[this.positionInCache]);
            return this.cachedBreakPositions[this.positionInCache];
        }
        this.cachedBreakPositions = null;
        int previous = super.previous();
        if (this.cachedBreakPositions != null) {
            this.positionInCache = this.cachedBreakPositions.length - 2;
        }
        return previous;
    }
    
    @Override
    public int preceding(int n) {
        CharacterIterator text = this.getText();
        RuleBasedBreakIterator.checkOffset(n, text);
        if (this.cachedBreakPositions == null || n <= this.cachedBreakPositions[0] || n > this.cachedBreakPositions[this.cachedBreakPositions.length - 1]) {
            this.cachedBreakPositions = null;
            return super.preceding(n);
        }
        this.positionInCache = 0;
        while (this.positionInCache < this.cachedBreakPositions.length && n > this.cachedBreakPositions[this.positionInCache]) {
            ++this.positionInCache;
        }
        --this.positionInCache;
        text.setIndex(this.cachedBreakPositions[this.positionInCache]);
        return text.getIndex();
    }
    
    @Override
    public int following(int n) {
        CharacterIterator text = this.getText();
        RuleBasedBreakIterator.checkOffset(n, text);
        if (this.cachedBreakPositions == null || n < this.cachedBreakPositions[0] || n >= this.cachedBreakPositions[this.cachedBreakPositions.length - 1]) {
            this.cachedBreakPositions = null;
            return super.following(n);
        }
        this.positionInCache = 0;
        while (this.positionInCache < this.cachedBreakPositions.length && n >= this.cachedBreakPositions[this.positionInCache]) {
            ++this.positionInCache;
        }
        text.setIndex(this.cachedBreakPositions[this.positionInCache]);
        return text.getIndex();
    }
    
    @Override
    protected int handleNext() {
        CharacterIterator text = this.getText();
        if (this.cachedBreakPositions == null || this.positionInCache == this.cachedBreakPositions.length - 1) {
            int index = text.getIndex();
            this.dictionaryCharCount = 0;
            int handleNext = super.handleNext();
            if (this.dictionaryCharCount > 1 && handleNext - index > 1) {
                this.divideUpDictionaryRange(index, handleNext);
            }
            else {
                this.cachedBreakPositions = null;
                return handleNext;
            }
        }
        if (this.cachedBreakPositions != null) {
            ++this.positionInCache;
            text.setIndex(this.cachedBreakPositions[this.positionInCache]);
            return this.cachedBreakPositions[this.positionInCache];
        }
        return -9999;
    }
    
    @Override
    protected int lookupCategory(int n) {
        int lookupCategory = super.lookupCategory(n);
        if (lookupCategory != -1 && this.categoryFlags[lookupCategory]) {
            ++this.dictionaryCharCount;
        }
        return lookupCategory;
    }
    
    private void divideUpDictionaryRange(int index, int i) {
        CharacterIterator text = this.getText();
        text.setIndex(index);
        for (int n = this.lookupCategory(this.getCurrent()); n == -1 || !this.categoryFlags[n]; n = this.lookupCategory(this.getNext())) {}
        Stack o = new Stack();
        Stack o2 = new Stack();
        ArrayList o3 = new ArrayList();
        int nextStateFromCharacter = 0;
        int n2 = text.getIndex();
        Stack<Object> stack = null;
        int n3 = this.getCurrent();
        while (true) {
            if (this.dictionary.getNextState(nextStateFromCharacter, 0) == -1) {
                o2.push(Integer.valueOf(text.getIndex()));
            }
            nextStateFromCharacter = this.dictionary.getNextStateFromCharacter(nextStateFromCharacter, n3);
            if (nextStateFromCharacter == -1) {
                o.push(Integer.valueOf(text.getIndex()));
                break;
            }
            if (nextStateFromCharacter == 0 || text.getIndex() >= i) {
                if (text.getIndex() > n2) {
                    n2 = text.getIndex();
                    stack = (Stack)o.clone();
                }
                while (!o2.isEmpty() && o3.contains(o2.peek())) {
                    o2.pop();
                }
                if (o2.isEmpty()) {
                    if (stack != null) {
                        o = stack;
                        if (n2 < i) {
                            text.setIndex(n2 + 1);
                        }
                        else {
                            break;
                        }
                    }
                    else {
                        if ((o.size() == 0 || (int)o.peek() != text.getIndex()) && text.getIndex() != index) {
                            o.push(new Integer(text.getIndex()));
                        }
                        this.getNext();
                        o.push(new Integer(text.getIndex()));
                    }
                }
                else {
                    Integer item = (Integer)o2.pop();
                    while (!o.isEmpty() && (int)item < (int)o.peek()) {
                        o3.add((Object)o.pop());
                    }
                    o.push(item);
                    text.setIndex((int)o.peek());
                }
                n3 = this.getCurrent();
                if (text.getIndex() < i) {
                    continue;
                }
                break;
            }
            else {
                n3 = this.getNext();
            }
        }
        if (!o.isEmpty()) {
            o.pop();
        }
        o.push(Integer.valueOf(i));
        (this.cachedBreakPositions = new int[o.size() + 1])[0] = index;
        for (int j = 0; j < o.size(); ++j) {
            this.cachedBreakPositions[j + 1] = (int)o.elementAt(j);
        }
        this.positionInCache = 0;
    }
}
